@@ -1404,7 +1404,7 @@ FFN(siLroom_report)(void) {
   for (i = 0;  i < (int)t_end;  i++) {
     if (tm_table[i].tm_type == (enum type)i) {
       vs_check_push(make_fixnum(TM_NUSED(tm_table[i])));
-      vs_push(make_fixnum(tm_table[i].tm_nfree));
+      vs_push(make_fixnum(tm_table[i].tm_nfree+tm_table[i].tm_alt_nfree));
       vs_push(make_fixnum(tm_table[i].tm_npage));
       vs_push(make_fixnum(tm_table[i].tm_maxpage));
       vs_push(make_fixnum(tm_table[i].tm_gbccount));
@@ -1426,7 +1426,7 @@ FFN(siLreset_gbc_count)(void) {
   check_arg(0);
   
   for (i = 0;  i < (int)t_other;  i++)
-    tm_table[i].tm_gbccount = tm_table[i].tm_adjgbccnt = 0;
+    tm_table[i].tm_gbccount = tm_table[i].tm_adjgbccnt = tm_table[i].tm_opt_maxpage = 0;
 }
 
 /* copy S bytes starting at P to beyond rb_pointer1 (temporarily)
