@@ -58,7 +58,11 @@ void initialize_process();
 #endif
 
 #ifdef _WIN32
+
+extern void init_shared_memory (void);
+
 #  include <fcntl.h>
+
 #endif
 
 #define LISP_IMPLEMENTATION_VERSION "April 1994"
@@ -72,16 +76,10 @@ int page_multiple        = 1;
 
 int    debug             = FALSE;		/* debug switch */
 int    initflag          = FALSE;		/* initialized flag */
+int    raw_image         = FALSE;		/* raw or saved image */
 
 int    stack_multiple    = 1;
-int    cssize            = 0;
-
-#ifdef _WIN32
-unsigned int _dbegin     = 0x10100000;
-#  ifndef AV
-unsigned int _stacktop, _stackbottom;
-#  endif
-#endif
+unsigned int cssize      = 0;
 
 extern bool saving_system;
 extern long real_maxpage;
@@ -116,6 +114,7 @@ main(int argc, char **argv, char **envp) {
 #if defined ( BSD ) && defined ( RLIMIT_STACK )
     struct rlimit rl;
 #endif
+    set_maxpage();
 
 #ifdef NEED_NONRANDOM_SBRK
 #if SIZEOF_LONG == 4
@@ -130,25 +129,18 @@ main(int argc, char **argv, char **envp) {
     extern void init_darwin_zone_compat ();
     init_darwin_zone_compat ();
 #endif
-
 #ifdef RECREATE_HEAP
     RECREATE_HEAP
 #endif
-	
 #if defined ( _WIN32 ) && !defined ( AV )
-    {
-        unsigned int dummy;
-	    
-        _stackbottom = (unsigned int ) &argc;
-        _stacktop    = _stackbottom - CSSIZE; // ???
-
-    }
-#endif
+#  error BLAH
+#endif        
+	
     setbuf(stdin, stdin_buf); 
     setbuf(stdout, stdout_buf);
 #ifdef _WIN32
     _fmode = _O_BINARY;
-    _setmode( _fileno( stdin ), _O_BINARY );
+    _setmode( _fileno( stdin ),  _O_BINARY );
     _setmode( _fileno( stdout ), _O_BINARY );
     _setmode( _fileno( stderr ), _O_BINARY );
 #endif
@@ -221,12 +213,13 @@ main(int argc, char **argv, char **envp) {
     }
 
     vs_top = vs_base = vs_org;
-    clear_stack ( vs_top, vs_top+200 );
+    clear_stack ( vs_top, vs_limit );
     ihs_top = ihs_org-1;
     bds_top = bds_org-1;
     frs_top = frs_org-1;
     cs_org = &argc;
-    cssize = CSSIZE;
+    /* CSSIZE in bytes, cssize for pointer arithmetic. */ 
+    cssize = CSSIZE / sizeof ( void * );
     install_segmentation_catcher();
 
 #ifdef BSD
@@ -264,10 +257,10 @@ main(int argc, char **argv, char **envp) {
 #endif /* BSD */
 
 #ifdef AV
+    
     cs_limit = cs_org - cssize;
 #endif
 
-    set_maxpage();
 
 #ifdef SETUP_SIG_STACK
     SETUP_SIG_STACK
@@ -293,6 +286,10 @@ main(int argc, char **argv, char **envp) {
 #ifdef SGC
     memprotect_test_reset();
 #endif
+#ifdef GCL_GPROF
+	if (atexit(gprof_cleanup))
+	  error("Cannot setup gprof_cleanup on exit");
+#endif
 
     if (initflag) {
         if (saving_system) {
@@ -368,6 +365,8 @@ main(int argc, char **argv, char **envp) {
 
     interrupt_enable = TRUE;
 
+    raw_image=TRUE;
+ 
     super_funcall(sStop_level);
 
     return 0;
@@ -707,10 +706,10 @@ FFN(siLcatch_fatal)(int i) {
 static void
 reset_cstack_limit(int arg) {
 #ifdef AV
-  if (&arg > cs_org - cssize + 16)
-    cs_limit = cs_org - cssize;
+  if ( &arg > cs_org - cssize + 16 )
+      cs_limit = cs_org - cssize;
   else
-    error("can't reset cs_limit");
+      error ( "can't reset cs_limit" );
 #endif
 }
 
@@ -803,7 +802,21 @@ FFN(siLaddress)(void) {
 static void
 FFN(siLnani)(void) {
   check_arg(1);
-  vs_base[0] = (object)fixint(vs_base[0]);
+  switch(type_of(vs_base[0])) {
+  case t_fixnum:
+    vs_base[0] = (object)fixint(vs_base[0]);
+    break;
+  case t_bignum:
+    if (vs_base[0]->big.big_mpz_t._mp_size!=1)
+      FEerror("Integer too big",0);
+    /*FIXME 64*/
+    fprintf(stderr,"Warning, bignum in nani\n");
+    vs_base[0]=(object)*vs_base[0]->big.big_mpz_t._mp_d;
+    break;
+  default:
+    FEerror("Not an integer",0);
+    break;
+  }    
 }
 
 static void
@@ -863,8 +876,10 @@ FFN(siLsave_system)(void) {
     saving_system = TRUE;
   GBC(t_contiguous);
   
-  
-  
+#ifdef GCL_GPROF
+  gprof_cleanup();
+#endif
+    
 #if defined(BSD) || defined(ATT)  
   brk(core_end);
   /* printf( "(breaking at core_end = %x in main ,)",core_end); */
@@ -877,10 +892,10 @@ FFN(siLsave_system)(void) {
 /*  #ifdef AOSVS */
   
 /*  #endif */
-  cbgbccount = tm_table[t_contiguous].tm_adjgbccnt = 0;
-  rbgbccount = tm_table[t_relocatable].tm_adjgbccnt = 0;
+  cbgbccount = tm_table[t_contiguous].tm_adjgbccnt = tm_table[t_contiguous].tm_opt_maxpage = 0;
+  rbgbccount = tm_table[t_relocatable].tm_adjgbccnt = tm_table[t_relocatable].tm_opt_maxpage = 0;
   for (i = 0;  i < (int)t_end;  i++)
-    tm_table[i].tm_gbccount = tm_table[i].tm_adjgbccnt = 0;
+    tm_table[i].tm_gbccount = tm_table[i].tm_adjgbccnt = tm_table[i].tm_opt_maxpage = 0;
   Lsave();
   saving_system = FALSE;
   alloc_page(-(holepage+nrbpage));
@@ -935,6 +950,9 @@ init_main(void) {
 #ifdef GMP
   ADD_FEATURE("GMP");
 #endif	 
+#ifdef GCL_GPROF
+  ADD_FEATURE("GPROF");
+#endif	 
   
 #if defined ( UNIX ) && !defined ( _WIN32 )
   ADD_FEATURE("UNIX");
@@ -992,8 +1010,11 @@ init_main(void) {
 #ifdef HAVE_READLINE
   ADD_FEATURE("READLINE");
 #endif
-#if defined(HAVE_LIBBFD) && !defined(USE_DLOPEN)
+#if !defined(USE_DLOPEN)
+  ADD_FEATURE("NATIVE-RELOC");
+#if defined(HAVE_LIBBFD) 
   ADD_FEATURE("BFD");
+#endif
 #endif
   ADD_FEATURE("UNEXEC");
 
